@@ -1,5 +1,6 @@
+from abc import abstractmethod
 from dataclasses import dataclass, asdict, field
-from typing import Tuple, Union, Literal, Mapping, Any, Optional, Dict
+from typing import Tuple, Union, Literal, Mapping, Any, Optional, Dict, Callable
 
 import numpy as np
 import rioxarray  # noqa # pylint: disable=unused-import
@@ -7,11 +8,12 @@ import xarray as xr
 from affine import Affine
 from numpy.typing import NDArray, DTypeLike
 from xarray import DataArray, Dataset
+from yaml import YAMLObject, SafeLoader
 
 from eotransform_xarray.storage.storage import Storage
 
 try:
-    from numba import njit, prange
+    from numba import njit, prange, guvectorize, float64, float32
     from pyresample import SwathDefinition, AreaDefinition
     from pyresample.kd_tree import get_neighbour_info
 except ImportError:
@@ -20,6 +22,9 @@ except ImportError:
 
 from eotransform_xarray.transformers import TransformerOfDataArray
 
+EngineType = Literal['dask', 'numba']
+EmptyRasterFactory = Callable[[Tuple[int, ...], DTypeLike], NDArray]
+
 
 @dataclass
 class Swath:
@@ -89,16 +94,44 @@ class StorageIntoTheVoid(Storage):
         pass
 
 
+class EngineConfig(YAMLObject):
+    @property
+    @abstractmethod
+    def type(self) -> EngineType:
+        ...
+
+
 @dataclass
-class DaskConfig:
+class DaskConfig(EngineConfig):
+    yaml_tag = "!ResampleDaskConfig"
+    yaml_loader = SafeLoader
+
     raster_chunk_sizes: Tuple[int, int]
 
+    @property
+    def type(self) -> EngineType:
+        return "dask"
+
 
 @dataclass
-class ProcessingConfig:
+class NumbaConfig(EngineConfig):
+    yaml_tag = "!ResampleNumbaConfig"
+    yaml_loader = SafeLoader
+
+    @property
+    def type(self) -> EngineType:
+        return "numba"
+
+
+@dataclass
+class ProcessingConfig(YAMLObject):
+    yaml_tag = "!ResampleProcessing"
+    yaml_loader = SafeLoader
+
     num_parameter_calc_procs: int = 1
+    num_lookup_segments: Optional[int] = None
     parameter_storage: Storage = field(default_factory=StorageIntoTheVoid)
-    resampling_engine: Optional[Union[Literal['numba'], DaskConfig]] = 'numba'
+    resampling_engine: Optional[Union[NumbaConfig, DaskConfig]] = field(default_factory=NumbaConfig)
     load_in_resampling_params: Optional[Dict] = None
     load_out_resampling_params: Optional[Dict] = None
 
@@ -108,9 +141,11 @@ class ResampleWithGauss(TransformerOfDataArray):
         ...
 
     def __init__(self, swath_src: Swath, area_dst: Area, sigma: float, neighbours: int, lookup_radius: float,
-                 processing_config: Optional[ProcessingConfig] = None):
+                 processing_config: Optional[ProcessingConfig] = None,
+                 empty_out_raster_factory: Optional[EmptyRasterFactory] = None):
         self._area_dst = area_dst
         self._proc_cfg = processing_config or ProcessingConfig()
+        self._empty_out_raster_factory = empty_out_raster_factory or make_empty_raster
         if self._proc_cfg.parameter_storage.exists():
             self._projection_params = ProjectionParameter.from_storage(self._proc_cfg.parameter_storage)
         else:
@@ -132,7 +167,8 @@ class ResampleWithGauss(TransformerOfDataArray):
         ar_def = AreaDefinition(area.name, area.description, "proj_id", area.projection, area.columns, area.rows,
                                 area.extent.to_tuple())
         val_in_idc, val_out_idc, idc, distances = get_neighbour_info(sw_def, ar_def, lookup_radius, neighbours,
-                                                                     nprocs=self._proc_cfg.num_parameter_calc_procs)
+                                                                     nprocs=self._proc_cfg.num_parameter_calc_procs,
+                                                                     segments=self._proc_cfg.num_lookup_segments)
         packed_idc = MaybePacked(idc) | np.uint8 | np.uint16 | np.uint32 | np.uint64
         packed_idc = packed_idc.value.reshape((area.rows, area.columns, -1))
         distances = distances.astype(np.float32).reshape((area.rows, area.columns, -1,))
@@ -146,17 +182,20 @@ class ResampleWithGauss(TransformerOfDataArray):
                                   'mask': (('y', 'x', 'cell'), out_mask)}) \
             .rio.write_crs(area.projection).rio.write_transform(area.transform)
 
-        if self._proc_cfg.resampling_engine == 'numba':
+        if self._proc_cfg.resampling_engine.type == 'numba':
             out_resampling = out_resampling.chunk({'neighbours': -1, 'cell': -1, 'y': -1, 'x': -1})
         else:
             rc = self._proc_cfg.resampling_engine.raster_chunk_sizes
             out_resampling = out_resampling.chunk({'neighbours': -1, 'cell': -1, 'y': rc[0], 'x': rc[1]})
         return ProjectionParameter(in_resampling, out_resampling)
 
-    @staticmethod
-    def _distances_to_gauss_weights(distances: DataArray, sigma: float) -> DataArray:
-        sig_sqrd = sigma ** 2
-        return np.exp(-distances ** 2 / sig_sqrd)
+    def _distances_to_gauss_weights(self, distances: DataArray, sigma: float) -> DataArray:
+        if self._proc_cfg.resampling_engine.type == 'numba':
+            distances.values = _distance_to_gauss_weight(distances.values, sigma ** 2)
+            return distances
+        else:
+            sig_sqrd = sigma ** 2
+            return np.exp(-distances ** 2 / sig_sqrd)
 
     def __call__(self, x: DataArray) -> DataArray:
         self._sanity_check_input(x)
@@ -167,14 +206,16 @@ class ResampleWithGauss(TransformerOfDataArray):
         out_valid = self._projection_params.out_resampling['mask'][:, :, 0].astype(bool)
         if self._proc_cfg.resampling_engine == 'numba':
             resampled = DataArray(
-                _resample_numba(x.values, indices.values, weights.values, out_valid.values),
+                _resample_numba(x.values, indices.values, weights.values, out_valid.values,
+                                self._empty_out_raster_factory),
                 dims=x.dims[:2] + out_valid.dims,
                 coords={**{k: v for k, v in x.coords.items() if k in x.dims[:2]}, **out_valid.coords})
         else:
             resampled = xr.apply_ufunc(_resample_dask, x, indices, weights, out_valid,
                                        input_core_dims=[x.dims[-1:], ['neighbours'], ['neighbours'], []],
                                        output_dtypes=[x.dtype],
-                                       dask='parallelized', keep_attrs=True)
+                                       dask='parallelized', keep_attrs=True,
+                                       kwargs=dict(out_fac=self._empty_out_raster_factory))
         resampled.attrs = x.attrs
         return resampled
 
@@ -185,17 +226,29 @@ class ResampleWithGauss(TransformerOfDataArray):
                                                   f"{self._projection_params.in_resampling.sizes} != {x.shape}")
 
 
-def _resample_dask(in_data: NDArray, indices: NDArray, weights: NDArray, out_valid: NDArray) -> NDArray:
+def make_empty_raster(requested_shape: Tuple[int, ...], requested_dtype: DTypeLike) -> NDArray:
+    return np.empty(requested_shape, dtype=requested_dtype)
+
+
+@guvectorize([(float32[:], float32, float32[:]),
+              (float64[:], float64, float64[:])], '(),()->()', target='parallel', nopython=True)
+def _distance_to_gauss_weight(distance, sigma_sqrd, out):
+    out[0] = np.exp(-distance[0] ** 2 / sigma_sqrd)
+
+
+def _resample_dask(in_data: NDArray, indices: NDArray, weights: NDArray, out_valid: NDArray,
+                   out_fac: EmptyRasterFactory) -> NDArray:
     in_data = in_data.squeeze((2, 3))
     times, parameters = in_data.shape[:2]
-    out = np.full((times, parameters) + out_valid.shape, np.nan, dtype=in_data.dtype)
+    out = out_fac((times, parameters) + out_valid.shape, in_data.dtype)
     _resample_to_single_threaded(in_data, indices, weights, out_valid, out)
     return out
 
 
-def _resample_numba(in_data: NDArray, indices: NDArray, weights: NDArray, out_valid: NDArray) -> NDArray:
+def _resample_numba(in_data: NDArray, indices: NDArray, weights: NDArray, out_valid: NDArray,
+                    out_fac: EmptyRasterFactory) -> NDArray:
     times, parameters = in_data.shape[:2]
-    out = np.full((times, parameters) + out_valid.shape, np.nan, dtype=in_data.dtype)
+    out = out_fac((times, parameters) + out_valid.shape, in_data.dtype)
     _resample_to_parallel(in_data, indices, weights, out_valid, out)
     return out
 
